import logging

import numpy as np
from scipy.special import gamma


class AnalyticalGammaToGumbel:
    """
    Compute relevant quantities related to the Gumbel distribution arising from the
    maximum over an ensemble of iid Gamma-distributed random variables (which includes
    the chi-squared distribution as a special case).
    """

    def __init__(self, dofs=None, shape=None, scale=None, rate=None):
        """
        The single-variable probability distribution can either be described as a
        chi-squared distribution with `dofs` degrees of freedom,
        a Gamma distribution with shape and scale parameters (`shape`, `scale`),
        or a Gamma distribution with shape and rate parameters (`shape`, `rate`).

        A chi-squared distribution with `dofs` degrees of freedom corresponds to a
        Gamma distribution with `shape=dofs/2` and `scale=2`, `rate=1/2`. 

        Parameters
        ----------
        dofs: float, optional
            Degrees of freedom of the underlying chi-squared distribution.
            Authomatically implies `shape=dofs/2` and `scale=2`.
            This parameter is incompatible with `shape`, `scale`, and `rate`.
        shape: float, optional
            Shape parameter of the underlying Gamma distribution.
            This parameter must be given jointly with `scale` or `rate`
            and is incompatible with `dofs`.
        scale: float, optional
            Scale parameter of the underlying gamma distribution 
            (inverse of the rate parameter).
            This parameter must be given jointly with `shape` and is
            incompatible with `dofs` or `rate`.
        rate: float, optional
            Rate parameter of the underlying gamma distribution 
            (inverse of the scale parameter).
            This parameter must be given jointly with `shape` and is
            incompatible with `dofs` or `scale`.
        """
        
        # Check the validity of the input
        if dofs:
            if shape:
                raise ValueError("Only one out of dofs and shape can be given")
            shape = dofs / 2
            scale = 2.
        elif shape:
            if not (bool(scale) ^ bool(rate)):
                raise ValueError("Only one out of scale and rate can be given")
            elif rate:
                scale = 1 / rate
        else:
            raise ValueError("Invalid input parameters: "
                    "Please enter dofs, (scale, shape) or (scale, rate)")
    
        self._set_shape_scale(shape=shape, scale=scale)

    def _set_shape_scale(self, shape, scale):
        self.shape = shape
        self.scale = scale

    def get_gumbel_loc_scale(self, n, improved=True):
        """
        Compute the theoretical estimation of the location and scale parameters of the
        Gumbel distribution resulting from maximizing over a batch of `n` independent
        random variables following the distribution specified by this class.

        We provide two implementations of said parameters, selected by
        switching the `improved` flag.

        If improved is falsy, location and scale parameters are taken from table 3.4.4
        in "Modelling Extremal Events for Insurance and Finance", P. Embrechts,
        C. Klüppelberg, T. Mikosch (1997) https://doi.org/10.1007/978-3-642-33483-2.

        If improved is truthy (recommended), location and scale parameters are taken
        from Eqs. (32) and (33) in  Gasull, López-Salcedo, and Utzet, TEST 24, 714–733
        (2015) https://doi.org/10.1007/s11749-015-0431-9 .

        The improved option returns better estimates for n < 1e6 (and possibly beyond).

        Parameters
        ----------
        n: int
            Number of random variables within the batch being maximized.
        improved: bool
            If True (recommended), apply the improved loc and scale parameters. See
            the main docstring for details.

        Returns
        -------+
        loc: float
            Location parameter of the resulting Gumbel distribution.
        scale: float
            Scale parameter of the resulting Gumbel distribution.
        """

        logn = np.log(n)
        gamma_shape = gamma(self.shape)

        if improved:
            logn_gamma = np.log(n / gamma_shape)
            shape_m1 = self.shape - 1
            Bn = logn_gamma + shape_m1 * np.log(shape_m1)
            logBn = np.log(Bn)

            loc = self.scale * (
                logn_gamma
                + shape_m1 * logBn
                + (shape_m1 * shape_m1 * (logBn - np.log(shape_m1)) + shape_m1) / Bn
            )
            scale = (self.scale * loc * (self.scale * shape_m1 + loc)) / (
                loc * loc - self.scale * self.scale * shape_m1 * (self.shape - 2)
            )

        else:
            loc = self.scale * (
                logn + (self.shape - 1) * np.log(logn) + np.log(gamma_shape)
            )
            scale = self.scale

        return loc, scale

    def get_gumbel_mean_std(self, n, improved=True):
        """
        Compute the theoretical estimation of the mean and standard deviation parameters
        of the Gumbel distribution resulting from maximizing over a batch of `n` independent
        random variables following the distribution specified by this class.

        These results computed from the `loc` and `scale` parameters returned by
        `self.gumbel_mean_std`. See the docstring therein for a further explanation of the
        parameters.

        Parameters
        ----------
        n: int
            Number of random variables within the batch being maximized.
        improved: bool
            If True (recommended), apply the improved loc and scale parameters. See
            the main docstring for details.

        Returns
        -------
        mean: float
            Mean of the Gumbel distribution, computed as `loc + euler_gamma * scale`,
            where `euler_gamma=0.5772...` represents the Euler-Mascheroni constant
            (sequence A001620 in the OEIS).
        std: float
            Standard deviation of the Gumbel distribution,
            computed as `pi/sqrt(6) * scale`.
        """
        loc, scale = self.get_gumbel_loc_scale(n=n, improved=improved)
        mean = loc + np.euler_gamma * scale
        std = np.pi / np.sqrt(6) * scale
        return mean, std



import logging
import numpy as np
from scipy import signal, stats
from skimage.filters import threshold_minimum
from tqdm import tqdm, trange


class BatchMaxGumbel:
    def __init__(self, samples, batch_size=None, shuffle=True, shuffle_seed=None):
        """
        Basic implementation of the method described in [0].
        Batch a given set of samples, retrieve batch-wise maxima and fit a Gumbel distribution.

        [0] R. Tenorio, L. M. Modafferi, D. Keitel, A. M. Sintes

        Parameters
        ----------
        samples: array-like 1D or 2D
            Samples to which a Gumbel distribution will be fitted.
            If 2D, samples should be located along the last column (-1 on the last index).

        batch_size: int
            Number of samples per batch.

        shuffle: bool
            Whether to shuffle the samples before batching (recommended) or not.
            Shuffling uses numpy's permutation method from numpy.random.default_rng.

        shuffle_seed: int, optional
            Seed to initialize numpy.random.default_rng
        """

        dimension = getattr(samples, "ndim", 0)
        if dimension == 2:
            samples = samples[:, -1]
        elif dimension > 2:
            raise ValueError(
                "Input samples array is > 2D. "
                "Please, slice out samples before using this class"
            )

        logging.info("Starting distromax...")
        self.set_samples(samples)
        logging.info("Samples successfully set")
        self.batch_max_samples(batch_size, shuffle, shuffle_seed)
        logging.info("Batchmax samples successfully retrieved")
        self.fit()
        logging.info("Gumbel distribution successfully fitted")

    def set_samples(self, samples):

        self.samples = samples
        histogram, x = np.histogram(samples, density=True, bins="auto")
        self.dx = x[1] - x[0]
        self.x = 0.5 * (x[1:] + x[:-1])
        self.histogram = histogram

    def batch_max_samples(self, batch_size, shuffle, shuffle_seed):

        if not batch_size:
            self.batch_max = self.samples.copy()
        else:
            if shuffle:
                rng = np.random.default_rng(shuffle_seed)
                samples_to_batch = rng.permutation(self.samples)
            else:
                samples_to_batch = self.samples

            self.batch_max = np.maximum.reduceat(
                samples_to_batch, np.arange(0, len(samples_to_batch), batch_size)
            )

    def fit(self, samples=None):
        samples = samples if samples is not None else self.batch_max
        self.gumbel = stats.gumbel_r(*stats.gumbel_r.fit(samples))

    def max_propagation(self, num_batches):
        """
        Maximize the fitted Gumbel distribution over `num_batches` realizations
        using the analytical formula. The output is placed so as to be consistent
        with what `scipy.gumbel_r` expects.
        """
        loc, scale = self.gumbel.args
        propagated_params = (loc + scale * np.log(num_batches), scale)
        self.propagated_gumbel = stats.gumbel_r(*propagated_params)
        return propagated_params


class BatchMaxGumbelWithCleaning(BatchMaxGumbel):
    """
    Abstract implementation of the cleaning procedure described in Appendix B of [0].
    Parameter's meaning is implementation dependent.
    See specific implementations below.

    [0] R. Tenorio, L. M. Modafferi, D. Keitel, A. M. Sintes

    Parameters
    ----------
    num_iterations: int
        Number of times the notching procedure will be applied.
    stopping_det_stat: float
        If the detection statistic threshold scores below this value, no notching will take place.
    stopping_quantile: float
        Alternative way of specifying `stopping_det_stat` as a quantile of the detection statistic.
    threshold: callable or float
        Threshold to perform cleaning. By default, skimage.threshold_minim is used.
    """

    _num_iterations = 0
    _stopping_det_stat = None
    _stopping_quantile = None
    _threshold = None

    def __init__(
        self,
        samples,
        num_iterations=None,
        stopping_det_stat=None,
        stopping_quantile=None,
        threshold=None,
        **kwargs,
    ):

        if (
            not getattr(samples, "ndim", None)
            or (samples.ndim != 2)
            or (samples.shape[1] < 2)
        ):
            raise ValueError(
                "`samples` should be a 2D array containing frequency values in [:, 0] and "
                "corresponding statistic values in [:, -1]"
            )

        num_iterations = (
            num_iterations if num_iterations is not None else self._num_iterations
        )

        if (
            (threshold is not None)
            and (not callable(threshold))
            and (num_iterations > 1)
        ):
            logging.info(
                "Threshold was manually set to a value but `num_iterations > 1`. "
                "Setting num_iterations to 1."
            )
            num_iterations = 1

        self.notching_iterations = 0

        notched_samples = samples
        for iteration in trange(
            num_iterations,
            desc="Notching iterations",
            unit=" notches",
            dynamic_ncols=True,
        ):
            self.notching_iterations += 1
            notched_samples = self.notch_outliers(
                samples=notched_samples,
                stopping_det_stat=stopping_det_stat or self._stopping_det_stat,
                threshold=threshold or self._threshold,
                stopping_quantile=stopping_quantile or self._stopping_quantile,
            )
            if not notched_samples.size:
                raise RuntimeError(
                    "It appears the last notching iteration completely obliterated your samples.\n"
                    f"Current configuration is `num_iterations = {num_iterations}`, "
                    "you probably should reduce that number."
                )

        super().__init__(samples=notched_samples[:, -1], **kwargs)

    def notch_outliers(
        self,
        samples,
        stopping_quantile,
        stopping_det_stat,
        threshold,
    ):

        # Can't stress this enough: always sort by frequency
        samples = samples[samples[:, 0].argsort()]
        self.f0, self.unique_indices = np.unique(samples[:, 0], return_index=True)
        self.max_at_f0 = np.maximum.reduceat(samples[:, -1], self.unique_indices)

        self.stopping_quantile = stopping_quantile
        self.stopping_det_stat = stopping_det_stat

        if callable(threshold):
            self.threshold = threshold(self.max_at_f0)
        elif threshold is None:
            try:
                self.threshold = threshold_minimum(self.max_at_f0)
            except RuntimeError as e:
                tqdm.write(
                    f"Iteration {self.notching_iterations}: {e}, threshold goes to -np.inf"
                )
                self.threshold = -np.inf
        else:
            self.threshold = threshold

        return self.remove_samples(samples)


class BatchMaxGumbelNotchingOutliers(BatchMaxGumbelWithCleaning):
    """
    Basic implementation of the notching algorithm described in Appendix B of [0].

    This class operates on the distribution of the frequency-bin-wise maxima to compute thresholds
    and notch loud frequency bands.

    [0] R. Tenorio, L. M. Modafferi, D. Keitel, A. M. Sintes
    """

    _num_iterations = 5
    _stopping_quantile = 0.8

    def remove_samples(self, samples):

        self.stopping_det_stat = self.stopping_det_stat or np.quantile(
            self.max_at_f0, self.stopping_quantile
        )

        if self.threshold < self.stopping_det_stat:
            tqdm.write(
                f"Iteration {self.notching_iterations}: "
                f"Threshold is below {100 * self.stopping_quantile}% quantile, skipping notching..."
            )
            return samples

        notched_bands = self.max_at_f0 > self.threshold
        notched_bands_index = np.nonzero(notched_bands)[0]

        left_index_notch = notched_bands_index[
            np.diff(notched_bands_index, prepend=-1) > 1
        ]

        right_index_notch = (
            notched_bands_index[
                np.diff(notched_bands_index, append=notched_bands_index.max() + 2) > 1
            ]
            + 1
        )

        indices_to_delete = np.concatenate(
            [
                np.arange(self.unique_indices[left], self.unique_indices[right])
                for left, right in zip(left_index_notch, right_index_notch)
            ]
        )

        tqdm.write(
            f"Iteration {self.notching_iterations}: "
            f"Threshold={self.threshold:.2f}, "
            f"{100 * self.stopping_quantile}% foreground quantile at {self.stopping_det_stat:.2f}, "
            "notching {}/{} ({:.2f}%) samples in {}/{} ({:.2f}%) frequency bins".format(
                indices_to_delete.shape[0],
                samples.shape[0],
                100 * indices_to_delete.shape[0] / samples.shape[0],
                notched_bands.sum(),
                self.max_at_f0.shape[0],
                100 * notched_bands.sum() / self.max_at_f0.shape[0],
            )
        )

        return np.delete(samples, indices_to_delete, axis=0)
